

/* this file was generated by volk template utils, do not edit! */

/* -*- c++ -*- */
/*
 * Copyright 2011-2012 Free Software Foundation, Inc.
 *
 * This file is part of VOLK
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#ifndef INCLUDED_VOLK_CONFIG_FIXED_H
#define INCLUDED_VOLK_CONFIG_FIXED_H

#define LV_GENERIC 0
#define LV_SOFTFP 1
#define LV_HARDFP 2
#define LV_32 3
#define LV_64 4
#define LV_POPCOUNT 5
#define LV_MMX 6
#define LV_FMA 7
#define LV_SSE 8
#define LV_SSE2 9
#define LV_ORC 10
#define LV_NORC 11
#define LV_NEON 12
#define LV_NEONV7 13
#define LV_NEONV8 14
#define LV_SSE3 15
#define LV_SSSE3 16
#define LV_SSE4_A 17
#define LV_SSE4_1 18
#define LV_SSE4_2 19
#define LV_AVX 20
#define LV_AVX2 21
#define LV_AVX512F 22
#define LV_AVX512CD 23
#define LV_RISCV64 24

#endif /*INCLUDED_VOLK_CONFIG_FIXED*/
